from eth_utils import (
    denoms,
)

GAS_EXTCODEHASH_EIP1052 = 400
GAS_SSTORE_EIP1283_NOOP = 200
GAS_SSTORE_EIP1283_INIT = 20000
GAS_SSTORE_EIP1283_CLEAN = 5000
GAS_SSTORE_EIP1283_CLEAR_REFUND = 15000
GAS_SSTORE_EIP1283_RESET_CLEAR_REFUND = 19800
GAS_SSTORE_EIP1283_RESET_REFUND = 4800

EIP1234_BLOCK_REWARD = 2 * denoms.ether
