from eth_typing import (
    BlockNumber,
)

# https://github.com/ethereum/EIPs/blob/master/EIPS/eip-155.md
ROPSTEN_CHAIN_ID = 3

# Fork Blocks listed in ascending order

#
# DAO Block
#
DAO_FORK_ROPSTEN_BLOCK = BlockNumber(0)


#
# Tangerine Whistle Block
#
TANGERINE_WHISTLE_ROPSTEN_BLOCK = BlockNumber(0)


#
# Homestead Block
#
HOMESTEAD_ROPSTEN_BLOCK = BlockNumber(0)


#
# Spurious Dragon Block
#
SPURIOUS_DRAGON_ROPSTEN_BLOCK = BlockNumber(10)


#
# Byzantium Block
#
BYZANTIUM_ROPSTEN_BLOCK = BlockNumber(1700000)


#
# Constantinople
#
CONSTANTINOPLE_ROPSTEN_BLOCK = BlockNumber(4230000)


#
# Petersburg
#
PETERSBURG_ROPSTEN_BLOCK = BlockNumber(4939394)


#
# Istanbul
#
ISTANBUL_ROPSTEN_BLOCK = BlockNumber(6485846)

#
# Muir Glacier Block
#
MUIR_GLACIER_ROPSTEN_BLOCK = BlockNumber(7117117)
BERLIN_ROPSTEN_BLOCK = BlockNumber(9812189)
