from ..arrow_glacier.computation import (
    ArrowGlacierComputation,
)


class GrayGlacierComputation(ArrowGlacierComputation):
    """
    A class for all execution *message* computations in the ``GrayGlacier`` fork.
    Inherits from :class:`~eth.vm.forks.arrow_glacier.ArrowGlacierComputation`
    """

    pass
