from ..london.computation import (
    LondonComputation,
)


class ArrowGlacierComputation(LondonComputation):
    """
    A class for all execution *message* computations in the ``ArrowGlacier`` fork.
    Inherits from :class:`~eth.vm.forks.london.LondonComputation`
    """

    pass
