GAS_MOD_EXP_QUADRATIC_DENOMINATOR_EIP_2565 = 3

# EIP 2718
VALID_TRANSACTION_TYPES = range(0, 0x80)

# EIP 2929
WARM_STORAGE_READ_COST = 100
COLD_SLOAD_COST = 2100
COLD_ACCOUNT_ACCESS_COST = 2600

# EIP 2930
ACCESS_LIST_TRANSACTION_TYPE = 1
ACCESS_LIST_ADDRESS_COST_EIP_2930 = 2400
ACCESS_LIST_STORAGE_KEY_COST_EIP_2930 = 1900
