from .tangerine_whistle import (  # noqa: F401
    TangerineWhistleVM,
)
from .frontier import (  # noqa: F401
    FrontierVM,
)
from .homestead import (  # noqa: F401
    HomesteadVM,
)
from .spurious_dragon import (  # noqa: F401
    SpuriousDragonVM,
)
from .byzantium import (  # noqa: F401
    ByzantiumVM,
)
from .constantinople import (  # noqa: F401
    ConstantinopleVM,
)
from .petersburg import (  # noqa: F401
    PetersburgVM,
)
from .istanbul import (  # noqa: F401
    IstanbulVM,
)
from .muir_glacier import (  # noqa: F401
    MuirGlacierVM,
)
from .berlin import (  # noqa: F401
    BerlinVM,
)
from .london import (  # noqa: F401
    LondonVM,
)
from .arrow_glacier import (  # noqa: F401
    ArrowGlacierVM,
)
from .gray_glacier import (  # noqa: F401
    GrayGlacierVM,
)
from .paris import (  # noqa: F401
    ParisVM,
)
from .shanghai import (  # noqa: F401
    ShanghaiVM,
)
