from eth_typing import (
    BlockNumber,
)

GOERLI_CHAIN_ID = 5

# Fork Blocks listed in ascending order

#
# Petersburg
#
PETERSBURG_GOERLI_BLOCK = BlockNumber(0)

#
# Istanbul
#
ISTANBUL_GOERLI_BLOCK = BlockNumber(1561651)

BERLIN_GOERLI_BLOCK = BlockNumber(4460644)
