#
# EIP150 Mainnet Block
#
TANGERINE_WHISTLE_MAINNET_BLOCK = 2463000


#
# Gas Costs and Refunds
#
GAS_SELFDESTRUCT_EIP150 = 5000
GAS_CALL_EIP150 = 700
GAS_EXTCODE_EIP150 = 700
GAS_BALANCE_EIP150 = 400
GAS_SLOAD_EIP150 = 200
